@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014, 2015 netblue30 (netblue30@yahoo.com)
+ * Copyright (C) 2014-2016 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -21,178 +21,254 @@
 #include <sys/mount.h>
 #include <sys/stat.h>
 #include <linux/limits.h>
+#include <fnmatch.h>
 #include <glob.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <errno.h>
 
-// build /tmp/firejail directory
+static void create_empty_dir(void) {
+	struct stat s;
+	
+	if (stat(RUN_RO_DIR, &s)) {
+		/* coverity[toctou] */
+		int rv = mkdir(RUN_RO_DIR, S_IRUSR | S_IXUSR);
+		if (rv == -1)
+			errExit("mkdir");	
+		if (chown(RUN_RO_DIR, 0, 0) < 0)
+			errExit("chown");
+	}
+}
+
+static void create_empty_file(void) {
+	struct stat s;
+
+	if (stat(RUN_RO_FILE, &s)) {
+		/* coverity[toctou] */
+		FILE *fp = fopen(RUN_RO_FILE, "w");
+		if (!fp)
+			errExit("fopen");
+		fclose(fp);
+		if (chown(RUN_RO_FILE, 0, 0) < 0)
+			errExit("chown");
+		if (chmod(RUN_RO_FILE, S_IRUSR) < 0)
+			errExit("chown");
+	}
+}
+
+// build /run/firejail directory
 void fs_build_firejail_dir(void) {
 	struct stat s;
 
-	if (stat(FIREJAIL_DIR, &s)) {
+	if (stat(RUN_FIREJAIL_DIR, &s)) {
 		if (arg_debug)
-			printf("Creating %s directory\n", FIREJAIL_DIR);
+			printf("Creating %s directory\n", RUN_FIREJAIL_DIR);
 		/* coverity[toctou] */
-		int rv = mkdir(FIREJAIL_DIR, S_IRWXU | S_IRWXG | S_IRWXO);
+		int rv = mkdir(RUN_FIREJAIL_DIR, 0755);
 		if (rv == -1)
 			errExit("mkdir");
-		if (chown(FIREJAIL_DIR, 0, 0) < 0)
+		if (chown(RUN_FIREJAIL_DIR, 0, 0) < 0)
 			errExit("chown");
-		if (chmod(FIREJAIL_DIR, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
+		if (chmod(RUN_FIREJAIL_DIR, 0755) < 0)
 			errExit("chmod");
 	}
 	else { // check /tmp/firejail directory belongs to root end exit if doesn't!
 		if (s.st_uid != 0 || s.st_gid != 0) {
-			fprintf(stderr, "Error: non-root %s directory, exiting...\n", FIREJAIL_DIR);
+			fprintf(stderr, "Error: non-root %s directory, exiting...\n", RUN_FIREJAIL_DIR);
 			exit(1);
 		}
 	}
+	
+	create_empty_dir();
+	create_empty_file();
 }
 
 
 // build /tmp/firejail/mnt directory
 static int tmpfs_mounted = 0;
+#ifdef HAVE_CHROOT		
+static void fs_build_remount_mnt_dir(void) {
+	tmpfs_mounted = 0;
+	fs_build_mnt_dir();
+}
+#endif
+
 void fs_build_mnt_dir(void) {
 	struct stat s;
 	fs_build_firejail_dir();
 	
-	// create /tmp/firejail directory
-	if (stat(MNT_DIR, &s)) {
+	// create /run/firejail/mnt directory
+	if (stat(RUN_MNT_DIR, &s)) {
 		if (arg_debug)
-			printf("Creating %s directory\n", MNT_DIR);
+			printf("Creating %s directory\n", RUN_MNT_DIR);
 		/* coverity[toctou] */
-		int rv = mkdir(MNT_DIR, S_IRWXU | S_IRWXG | S_IRWXO);
+		int rv = mkdir(RUN_MNT_DIR, 0755);
 		if (rv == -1)
 			errExit("mkdir");
-		if (chown(MNT_DIR, 0, 0) < 0)
+		if (chown(RUN_MNT_DIR, 0, 0) < 0)
 			errExit("chown");
-		if (chmod(MNT_DIR, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
+		if (chmod(RUN_MNT_DIR, 0755) < 0)
 			errExit("chmod");
 	}
 
 	// ... and mount tmpfs on top of it
 	if (!tmpfs_mounted) {
-		// mount tmpfs on top of /tmp/firejail/mnt
+		// mount tmpfs on top of /run/firejail/mnt
 		if (arg_debug)
-			printf("Mounting tmpfs on %s directory\n", MNT_DIR);
-		if (mount("tmpfs", MNT_DIR, "tmpfs", MS_NOSUID | MS_STRICTATIME | MS_REC,  "mode=755,gid=0") < 0)
+			printf("Mounting tmpfs on %s directory\n", RUN_MNT_DIR);
+		if (mount("tmpfs", RUN_MNT_DIR, "tmpfs", MS_NOSUID | MS_STRICTATIME | MS_REC,  "mode=755,gid=0") < 0)
 			errExit("mounting /tmp/firejail/mnt");
 		tmpfs_mounted = 1;
 	}
 }
 
-// build /tmp/firejail/overlay directory
-void fs_build_overlay_dir(void) {
+// grab a copy of cp command
+void fs_build_cp_command(void) {
 	struct stat s;
-	fs_build_firejail_dir();
-	
-	// create /tmp/firejail directory
-	if (stat(OVERLAY_DIR, &s)) {
-		if (arg_debug)
-			printf("Creating %s directory\n", MNT_DIR);
+	fs_build_mnt_dir();
+	if (stat(RUN_CP_COMMAND, &s)) {
+		char* fname = realpath("/bin/cp", NULL);
+		if (fname == NULL) {
+			fprintf(stderr, "Error: /bin/cp not found\n");
+			exit(1);
+		}
+		if (stat(fname, &s)) {
+			fprintf(stderr, "Error: /bin/cp not found\n");
+			exit(1);
+		}
+		if (is_link(fname)) {
+			fprintf(stderr, "Error: invalid /bin/cp file\n");
+			exit(1);
+		}
+		int rv = copy_file(fname, RUN_CP_COMMAND);
+		if (rv) {
+			fprintf(stderr, "Error: cannot access /bin/cp\n");
+			exit(1);
+		}
 		/* coverity[toctou] */
-		int rv = mkdir(OVERLAY_DIR, S_IRWXU | S_IRWXG | S_IRWXO);
-		if (rv == -1)
-			errExit("mkdir");	
-		if (chown(OVERLAY_DIR, 0, 0) < 0)
+		if (chown(RUN_CP_COMMAND, 0, 0))
 			errExit("chown");
-		if (chmod(OVERLAY_DIR, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
+		if (chmod(RUN_CP_COMMAND, 0755))
 			errExit("chmod");
+			
+		free(fname);
 	}
 }
 
-
-
-
+// delete the temporary cp command
+void fs_delete_cp_command(void) {
+	unlink(RUN_CP_COMMAND);
+}
 
 //***********************************************
 // process profile file
 //***********************************************
 typedef enum {
 	BLACKLIST_FILE,
+	BLACKLIST_NOLOG,
 	MOUNT_READONLY,
 	MOUNT_TMPFS,
 	OPERATION_MAX
 } OPERATION;
 
+typedef enum {
+	UNSUCCESSFUL,
+	SUCCESSFUL
+} LAST_DISABLE_OPERATION;
+LAST_DISABLE_OPERATION last_disable = UNSUCCESSFUL;
 
-static char *create_empty_dir(void) {
-	struct stat s;
-	fs_build_firejail_dir();
-	
-	if (stat(RO_DIR, &s)) {
-		/* coverity[toctou] */
-		int rv = mkdir(RO_DIR, S_IRUSR | S_IXUSR);
-		if (rv == -1)
-			errExit("mkdir");	
-		if (chown(RO_DIR, 0, 0) < 0)
-			errExit("chown");
-	}
-	
-	return RO_DIR;
-}
-
-static char *create_empty_file(void) {
-	struct stat s;
-	fs_build_firejail_dir();
-
-	if (stat(RO_FILE, &s)) {
-		/* coverity[toctou] */
-		FILE *fp = fopen(RO_FILE, "w");
-		if (!fp)
-			errExit("fopen");
-		fclose(fp);
-		if (chown(RO_FILE, 0, 0) < 0)
-			errExit("chown");
-		if (chmod(RO_FILE, S_IRUSR) < 0)
-			errExit("chown");
-	}
-	
-	return RO_FILE;
-}
-
-static void disable_file(OPERATION op, const char *fname, const char *emptydir, const char *emptyfile) {
-	assert(fname);
-	assert(emptydir);
-	assert(emptyfile);
+static void disable_file(OPERATION op, const char *filename) {
+	assert(filename);
 	assert(op <OPERATION_MAX);
-
-	// if the file is a link, follow the link
-	char *lnk = NULL;
-	if (is_link(fname)) {
-		lnk = get_link(fname);
-		if (lnk)
-			fname = lnk;
-		else
-			fprintf(stderr, "Warning: cannot follow link %s, skipping...\n", fname);
+	last_disable = UNSUCCESSFUL;
+	
+	// rebuild /run/firejail directory in case tmpfs was mounted on top of /run
+	fs_build_firejail_dir();
+	
+	// Resolve all symlinks
+	char* fname = realpath(filename, NULL);
+	if (fname == NULL && errno != EACCES) {
+		if (arg_debug)
+			printf("Warning: %s is an invalid file, skipping...\n", filename);
+		return;
+	}
+	if (fname == NULL && errno == EACCES) {
+		if (arg_debug)
+			printf("Debug: no access to file %s, forcing mount\n", filename);
+		// realpath and stat funtions will fail on FUSE filesystems
+		// they don't seem to like a uid of 0
+		// force mounting
+		int rv = mount(RUN_RO_DIR, filename, "none", MS_BIND, "mode=400,gid=0");
+		if (rv == 0)
+			last_disable = SUCCESSFUL;
+		else {
+			rv = mount(RUN_RO_FILE, filename, "none", MS_BIND, "mode=400,gid=0");
+			if (rv == 0)
+				last_disable = SUCCESSFUL;
+		}
+		if (last_disable == SUCCESSFUL) {
+			if (arg_debug)
+				printf("Disable %s\n", filename);
+			if (op == BLACKLIST_FILE)
+				fs_logger2("blacklist", filename);
+			else
+				fs_logger2("blacklist-nolog", filename);
+		}
+		else {
+			if (arg_debug)
+				printf("Warning: %s is an invalid file, skipping...\n", filename);
+		}
+				
+		return;
 	}
 	
 	// if the file is not present, do nothing
 	struct stat s;
 	if (stat(fname, &s) == -1) {
-		if (lnk)
-			free(lnk);
+		if (arg_debug)
+			printf("Warning: %s does not exist, skipping...\n", fname);
+		free(fname);
 		return;
 	}
 
 	// modify the file
-	if (op == BLACKLIST_FILE) {
-		if (arg_debug)
-			printf("Disable %s\n", fname);
-		if (S_ISDIR(s.st_mode)) {
-			if (mount(emptydir, fname, "none", MS_BIND, "mode=400,gid=0") < 0)
-				errExit("disable file");
-		}
+	if (op == BLACKLIST_FILE || op == BLACKLIST_NOLOG) {
+		// some distros put all executables under /usr/bin and make /bin a symbolic link
+		if ((strcmp(fname, "/bin") == 0 || strcmp(fname, "/usr/bin") == 0) &&
+		      is_link(filename) &&
+		      S_ISDIR(s.st_mode))
+			fprintf(stderr, "Warning: %s directory link was not blacklisted\n", filename);
+			
 		else {
-			if (mount(emptyfile, fname, "none", MS_BIND, "mode=400,gid=0") < 0)
-				errExit("disable file");
+			if (arg_debug)
+				printf("Disable %s\n", fname);
+			else if (arg_debug_blacklists) {
+				printf("Disable %s", fname);
+				if (op == BLACKLIST_FILE)
+					printf("\n");
+				else
+					printf(" - no logging\n");
+			}
+			if (S_ISDIR(s.st_mode)) {
+				if (mount(RUN_RO_DIR, fname, "none", MS_BIND, "mode=400,gid=0") < 0)
+					errExit("disable file");
+			}
+			else {
+				if (mount(RUN_RO_FILE, fname, "none", MS_BIND, "mode=400,gid=0") < 0)
+					errExit("disable file");
+			}
+			last_disable = SUCCESSFUL;
+			if (op == BLACKLIST_FILE)
+				fs_logger2("blacklist", fname);
+			else
+				fs_logger2("blacklist-nolog", fname);
 		}
 	}
 	else if (op == MOUNT_READONLY) {
 		if (arg_debug)
 			printf("Mounting read-only %s\n", fname);
 		fs_rdonly(fname);
+// todo: last_disable = SUCCESSFUL;
 	}
 	else if (op == MOUNT_TMPFS) {
 		if (S_ISDIR(s.st_mode)) {
@@ -204,6 +280,8 @@ static void disable_file(OPERATION op, const char *fname, const char *emptydir,
 			/* coverity[toctou] */
 			if (chown(fname, s.st_uid, s.st_gid) == -1)
 				errExit("mounting tmpfs chmod");
+			last_disable = SUCCESSFUL;
+			fs_logger2("mount tmpfs on", fname);
 		}
 		else
 			printf("Warning: %s is not a directory; cannot mount a tmpfs on top of it.\n", fname);
@@ -211,56 +289,81 @@ static void disable_file(OPERATION op, const char *fname, const char *emptydir,
 	else
 		assert(0);
 
-	if (lnk)
-		free(lnk);
+	free(fname);
 }
 
-static void globbing(OPERATION op, const char *fname, const char *emptydir, const char *emptyfile) {
-	assert(fname);
-	assert(emptydir);
-	assert(emptyfile);
+// Treat pattern as a shell glob pattern and blacklist matching files
+static void globbing(OPERATION op, const char *pattern, const char *noblacklist[], size_t noblacklist_len) {
+	assert(pattern);
 
-	// filename globbing: expand * macro and continue processing for every single file
-	if (strchr(fname, '*')) {
-		glob_t globbuf;
-		globbuf.gl_offs = 0;
-		glob(fname, GLOB_DOOFFS, NULL, &globbuf);
-		int i;
-		for (i = 0; i < globbuf.gl_pathc; i++) {
-			assert(globbuf.gl_pathv[i]);
-			disable_file(op, globbuf.gl_pathv[i], emptydir, emptyfile);
-		}
+	glob_t globbuf;
+	// Profiles contain blacklists for files that might not exist on a user's machine.
+	// GLOB_NOCHECK makes that okay.
+	int globerr = glob(pattern, GLOB_NOCHECK | GLOB_NOSORT, NULL, &globbuf);
+	if (globerr) {
+		fprintf(stderr, "Error: failed to glob pattern %s\n", pattern);
+		exit(1);
 	}
-	else
-		disable_file(op, fname, emptydir, emptyfile);
+
+	size_t i, j;
+	for (i = 0; i < globbuf.gl_pathc; i++) {
+		char *path = globbuf.gl_pathv[i];
+		assert(path);
+		// /home/me/.* can glob to /home/me/.. which would blacklist /home/
+		const char *base = gnu_basename(path);
+		if (strcmp(base, ".") == 0 || strcmp(base, "..") == 0)
+			continue;
+		// noblacklist is expected to be short in normal cases, so stupid and correct brute force is okay
+		bool okay_to_blacklist = true;
+		for (j = 0; j < noblacklist_len; j++) {
+			int result = fnmatch(noblacklist[j], path, FNM_PATHNAME);
+			if (result == FNM_NOMATCH)
+				continue;
+			else if (result == 0) {
+				okay_to_blacklist = false;
+				break;
+			}
+			else {
+				fprintf(stderr, "Error: failed to compare path %s with pattern %s\n", path, noblacklist[j]);
+				exit(1);
+			}
+		}
+		if (okay_to_blacklist)
+			disable_file(op, path);
+		else if (arg_debug)
+			printf("Not blacklist %s\n", path);
+	}
+	globfree(&globbuf);
 }
 
-static void expand_path(OPERATION op, const char *path, const char *fname, const char *emptydir, const char *emptyfile) {
-	assert(path);
-	assert(fname);
-	assert(emptydir);
-	assert(emptyfile);
-	char newname[strlen(path) + strlen(fname) + 1];
-	sprintf(newname, "%s%s", path, fname);
-
-	globbing(op, newname, emptydir, emptyfile);
-}
 
 // blacklist files or directoies by mounting empty files on top of them
-void fs_blacklist(const char *homedir) {
+void fs_blacklist(void) {
+	char *homedir = cfg.homedir;
+	assert(homedir);
 	ProfileEntry *entry = cfg.profile;
 	if (!entry)
 		return;
 		
-	char *emptydir = create_empty_dir();
-	char *emptyfile = create_empty_file();
+	size_t noblacklist_c = 0;
+	size_t noblacklist_m = 32;
+	char **noblacklist = calloc(noblacklist_m, sizeof(*noblacklist));
+
+	if (noblacklist == NULL)
+		errExit("failed allocating memory for noblacklist entries");
 
 	while (entry) {
 		OPERATION op = OPERATION_MAX;
 		char *ptr;
 
-		// process blacklist command
-		if (strncmp(entry->data, "bind", 4) == 0)  {
+		// whitelist commands handled by fs_whitelist()
+		if (strncmp(entry->data, "whitelist ", 10) == 0 || *entry->data == '\0') {
+			entry = entry->next;
+			continue;
+		}
+
+		// process bind command
+		if (strncmp(entry->data, "bind ", 5) == 0)  {
 			char *dname1 = entry->data + 5;
 			char *dname2 = split_comma(dname1);
 			if (dname2 == NULL) {
@@ -297,16 +400,34 @@ void fs_blacklist(const char *homedir) {
 			continue;
 		}
 
+		// Process noblacklist command
+		if (strncmp(entry->data, "noblacklist ", 12) == 0) {
+			if (noblacklist_c >= noblacklist_m) {
+                                noblacklist_m *= 2;
+                                noblacklist = realloc(noblacklist, sizeof(*noblacklist) * noblacklist_m);
+                                if (noblacklist == NULL)
+                                        errExit("failed increasing memory for noblacklist entries");
+			}
+			else
+				noblacklist[noblacklist_c++] = expand_home(entry->data + 12, homedir);
+			entry = entry->next;
+			continue;
+		}
+
 		// process blacklist command
-		if (strncmp(entry->data, "blacklist", 9) == 0)  {
+		if (strncmp(entry->data, "blacklist ", 10) == 0)  {
 			ptr = entry->data + 10;
 			op = BLACKLIST_FILE;
 		}
-		else if (strncmp(entry->data, "read-only", 9) == 0) {
+		else if (strncmp(entry->data, "blacklist-nolog ", 16) == 0)  {
+			ptr = entry->data + 16;
+			op = BLACKLIST_NOLOG;
+		}
+		else if (strncmp(entry->data, "read-only ", 10) == 0) {
 			ptr = entry->data + 10;
 			op = MOUNT_READONLY;
 		}			
-		else if (strncmp(entry->data, "tmpfs", 5) == 0) {
+		else if (strncmp(entry->data, "tmpfs ", 6) == 0) {
 			ptr = entry->data + 6;
 			op = MOUNT_TMPFS;
 		}			
@@ -317,27 +438,35 @@ void fs_blacklist(const char *homedir) {
 		}
 
 		// replace home macro in blacklist array
-		char *new_name = NULL;
-		if (strncmp(ptr, "${HOME}", 7) == 0) {
-			if (asprintf(&new_name, "%s%s", homedir, ptr + 7) == -1)
-				errExit("asprintf");
-			ptr = new_name;
-		}
+		char *new_name = expand_home(ptr, homedir);
+		ptr = new_name;
 
 		// expand path macro - look for the file in /bin, /usr/bin, /sbin and  /usr/sbin directories
-		if (strncmp(ptr, "${PATH}", 7) == 0) {
-			expand_path(op, "/bin", ptr + 7, emptydir, emptyfile);
-			expand_path(op, "/sbin", ptr + 7, emptydir, emptyfile);
-			expand_path(op, "/usr/bin", ptr + 7, emptydir, emptyfile);
-			expand_path(op, "/usr/sbin", ptr + 7, emptydir, emptyfile);
+		if (ptr) {
+			if (strncmp(ptr, "${PATH}", 7) == 0) {
+				char *fname = ptr + 7;
+				size_t fname_len = strlen(fname);
+				char **path, *paths[] = {"/bin", "/sbin", "/usr/bin", "/usr/sbin", NULL};
+				for (path = &paths[0]; *path; path++) {
+					char newname[strlen(*path) + fname_len + 1];
+					sprintf(newname, "%s%s", *path, fname);
+					globbing(op, newname, (const char**)noblacklist, noblacklist_c);
+					if  (last_disable == SUCCESSFUL)
+						break;
+				}
+			}
+			else
+				globbing(op, ptr, (const char**)noblacklist, noblacklist_c);
 		}
-		else
-			globbing(op, ptr, emptydir, emptyfile);
 
 		if (new_name)
 			free(new_name);
 		entry = entry->next;
 	}
+
+	size_t i;
+	for (i = 0; i < noblacklist_c; i++) free(noblacklist[i]);
+        free(noblacklist);
 }
 
 //***********************************************
@@ -357,6 +486,7 @@ void fs_rdonly(const char *dir) {
 		// mount --bind -o remount,ro /bin
 		if (mount(NULL, dir, NULL, MS_BIND|MS_REMOUNT|MS_RDONLY|MS_REC, NULL) < 0)
 			errExit("mount read-only");
+		fs_logger2("read-only", dir);
 	}
 }
 void fs_rdonly_noexit(const char *dir) {
@@ -374,6 +504,8 @@ void fs_rdonly_noexit(const char *dir) {
 			merr = 1;
 		if (merr)
 			fprintf(stderr, "Warning: cannot mount %s read-only\n", dir); 
+		else
+			fs_logger2("read-only", dir);
 	}
 }
 
@@ -385,6 +517,7 @@ void fs_proc_sys_dev_boot(void) {
 		printf("Remounting /proc and /proc/sys filesystems\n");
 	if (mount("proc", "/proc", "proc", MS_NOSUID | MS_NOEXEC | MS_NODEV | MS_REC, NULL) < 0)
 		errExit("mounting /proc");
+	fs_logger("remount /proc");
 
 	// remount /proc/sys readonly
 	if (mount("/proc/sys", "/proc/sys", NULL, MS_BIND | MS_REC, NULL) < 0)
@@ -392,6 +525,7 @@ void fs_proc_sys_dev_boot(void) {
 
 	if (mount(NULL, "/proc/sys", NULL, MS_BIND | MS_REMOUNT | MS_RDONLY | MS_REC, NULL) < 0)
 		errExit("mounting /proc/sys");
+	fs_logger("read-only /proc/sys");
 
 
 	/* Mount a version of /sys that describes the network namespace */
@@ -401,53 +535,32 @@ void fs_proc_sys_dev_boot(void) {
 		fprintf(stderr, "Warning: failed to unmount /sys\n");
 	if (mount("sysfs", "/sys", "sysfs", MS_RDONLY|MS_NOSUID|MS_NOEXEC|MS_NODEV|MS_REC, NULL) < 0)
 		fprintf(stderr, "Warning: failed to mount /sys\n");
+	else
+		fs_logger("remount /sys");
+		
+	if (stat("/sys/firmware", &s) == 0) {
+		disable_file(BLACKLIST_FILE, "/sys/firmware");
+	}
+		
+	if (stat("/sys/hypervisor", &s) == 0) {
+		disable_file(BLACKLIST_FILE, "/sys/hypervisor");
+	}
+
+	if (stat("/sys/fs", &s) == 0) {
+		disable_file(BLACKLIST_FILE, "/sys/fs");
+	}
+
+	if (stat("/sys/module", &s) == 0) {
+		disable_file(BLACKLIST_FILE, "/sys/module");
+	}
+
+	if (stat("/sys/power", &s) == 0) {
+		disable_file(BLACKLIST_FILE, "/sys/power");
+	}
 
 //	if (mount("sysfs", "/sys", "sysfs", MS_RDONLY|MS_NOSUID|MS_NOEXEC|MS_NODEV|MS_REC, NULL) < 0)
 //		errExit("mounting /sys");
 
-
-	// mounting firejail kernel module files
-	if (stat("/proc/firejail-uptime", &s) == 0) {
-		errno = 0;
-		FILE *fp = fopen("/proc/firejail", "w");
-		int cnt = 0;
-		while (errno == EBUSY && cnt < 10) {
-			if (!fp) {
-				int s = random();
-				s /= 200000;
-				usleep(s);
-				fp = fopen("/proc/firejail", "w");
-			}
-			else
-				break;
-		}
-		if (!fp) {
-			fprintf(stderr, "Error: cannot register sandbox with firejail-lkm\n");
-			exit(1);
-		}	
-		if (fp) {
-			// registration
-			fprintf(fp, "register\n");
-			fflush(0);
-			// filtering x11 connect calls
-			if (arg_nox11) {
-				fprintf(fp, "no connect unix /tmp/.X11\n");
-				fflush(0);
-				printf("X11 access disabled\n");
-			}
-			if (arg_nodbus) {
-				fprintf(fp, "no connect unix /var/run/dbus/system_bus_socket\n");
-				fflush(0);
-				fprintf(fp, "no connect unix /tmp/dbus\n");
-				fflush(0);
-				printf("D-Bus access disabled\n");
-			}
-			fclose(fp);
-			if (mount("/proc/firejail-uptime", "/proc/uptime", NULL, MS_BIND|MS_REC, NULL) < 0)
-				fprintf(stderr, "Warning: cannot mount /proc/firejail-uptime\n");
-		}
-	}
-
 	// Disable SysRq
 	// a linux box can be shut down easily using the following commands (as root):
 	// # echo 1 > /proc/sys/kernel/sysrq
@@ -474,80 +587,80 @@ void fs_proc_sys_dev_boot(void) {
 	fs_rdonly_noexit("/proc/bus");
 	
 	// disable /proc/kcore
-	disable_file(BLACKLIST_FILE, "/proc/kcore", "not used", "/dev/null");
+	disable_file(BLACKLIST_FILE, "/proc/kcore");
 
 	// disable /proc/kallsyms
-	disable_file(BLACKLIST_FILE, "/proc/kallsyms", "not used", "/dev/null");
+	disable_file(BLACKLIST_FILE, "/proc/kallsyms");
 	
 	// disable /boot
 	if (stat("/boot", &s) == 0) {
 		if (arg_debug)
-			printf("Mounting a new /boot directory\n");
-		if (mount("tmpfs", "/boot", "tmpfs", MS_NOSUID | MS_NODEV | MS_STRICTATIME | MS_REC,  "mode=777,gid=0") < 0)
-			errExit("mounting /boot directory");
+			printf("Disable /boot directory\n");
+		disable_file(BLACKLIST_FILE, "/boot");
+	}
+	
+	// disable /selinux
+	if (stat("/selinux", &s) == 0) {
+		if (arg_debug)
+			printf("Disable /selinux directory\n");
+		disable_file(BLACKLIST_FILE, "/selinux");
 	}
 	
 	// disable /dev/port
 	if (stat("/dev/port", &s) == 0) {
-		disable_file(BLACKLIST_FILE, "/dev/port", "not used", "/dev/null");
+		disable_file(BLACKLIST_FILE, "/dev/port");
+	}
+	
+	if (getuid() != 0) {
+		// disable /dev/kmsg
+		if (stat("/dev/kmsg", &s) == 0) {
+			disable_file(BLACKLIST_FILE, "/dev/kmsg");
+		}
+		
+		// disable /proc/kmsg
+		if (stat("/proc/kmsg", &s) == 0) {
+			disable_file(BLACKLIST_FILE, "/proc/kmsg");
+		}
 	}
 }
 
-static void sanitize_home(void) {
-	// extract current /home directory data
-	struct dirent *dir;
-	DIR *d = opendir("/home");
-	if (d == NULL)
-		return;
+// disable firejail configuration in /etc/firejail and in ~/.config/firejail
+static void disable_firejail_config(void) {
+	struct stat s;
+	if (stat("/etc/firejail", &s) == 0)
+		disable_file(BLACKLIST_FILE, "/etc/firejail");
 
-	char *emptydir = create_empty_dir();
-	while ((dir = readdir(d))) {
-		if(strcmp(dir->d_name, "." ) == 0 || strcmp(dir->d_name, ".." ) == 0)
-			continue;
+	char *fname;
+	if (asprintf(&fname, "%s/.config/firejail", cfg.homedir) == -1)
+		errExit("asprintf");
+	if (stat(fname, &s) == 0)
+		disable_file(BLACKLIST_FILE, fname);
+	
+	if (stat("/usr/local/etc/firejail", &s) == 0)
+		disable_file(BLACKLIST_FILE, "/usr/local/etc/firejail");
 
-		if (dir->d_type == DT_DIR ) {
-			// get properties
-			struct stat s;
-			char *name;
-			if (asprintf(&name, "/home/%s", dir->d_name) == -1)
-				continue;
-			if (stat(name, &s) == -1)
-				continue;
-			if (S_ISLNK(s.st_mode)) {
-				free(name);
-				continue;
-			}
-			
-			if (strcmp(name, cfg.homedir) == 0)
-				continue;
-
-//			printf("directory %u %u:%u #%s#\n",
-//				s.st_mode,
-//				s.st_uid,
-//				s.st_gid,
-//				name);
-			
-			// disable directory
-			disable_file(BLACKLIST_FILE, name, emptydir, "not used");
-			free(name);
-		}			
+	if (strcmp(PREFIX, "/usr/local")) {
+		if (asprintf(&fname, "%s/etc/firejail", PREFIX) == -1)
+			errExit("asprintf");
+		if (stat(fname, &s) == 0)
+			disable_file(BLACKLIST_FILE, fname);
 	}
-	closedir(d);
+
+	
+	
+	free(fname);
 }
 
 
-
-
-
-
 // build a basic read-only filesystem
 void fs_basic_fs(void) {
 	if (arg_debug)
-		printf("Mounting read-only /bin, /sbin, /lib, /lib64, /usr, /etc, /var\n");
+		printf("Mounting read-only /bin, /sbin, /lib, /lib32, /lib64, /usr, /etc, /var\n");
 	fs_rdonly("/bin");
 	fs_rdonly("/sbin");
 	fs_rdonly("/lib");
 	fs_rdonly("/lib64");
+	fs_rdonly("/lib32");
 	fs_rdonly("/usr");
 	fs_rdonly("/etc");
 	fs_rdonly("/var");
@@ -562,9 +675,10 @@ void fs_basic_fs(void) {
 	fs_var_cache();
 	fs_var_utmp();
 
-	// only in user mode
-	if (getuid())
-		sanitize_home();
+	// don't leak user information
+	restrict_users();
+	
+	disable_firejail_config();
 }
 
 
@@ -624,40 +738,54 @@ void fs_overlayfs(void) {
 	fs_build_mnt_dir();
 
 	char *oroot;
-	if(asprintf(&oroot, "%s/oroot", MNT_DIR) == -1)
+	if(asprintf(&oroot, "%s/oroot", RUN_MNT_DIR) == -1)
 		errExit("asprintf");
-	if (mkdir(oroot, S_IRWXU | S_IRWXG | S_IRWXO))
+	if (mkdir(oroot, 0755))
 		errExit("mkdir");
 	if (chown(oroot, 0, 0) < 0)
 		errExit("chown");
-	if (chmod(oroot, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
+	if (chmod(oroot, 0755) < 0)
 		errExit("chmod");
 
+	char *basedir = RUN_MNT_DIR;
+	if (arg_overlay_keep) {
+		// set base for working and diff directories
+		basedir = cfg.overlay_dir;
+		if (mkdir(basedir, 0755) != 0) {
+			fprintf(stderr, "Error: cannot create overlay directory\n");
+			exit(1);
+		}
+	}
+
 	char *odiff;
-	if(asprintf(&odiff, "%s/odiff", MNT_DIR) == -1)
+	if(asprintf(&odiff, "%s/odiff", basedir) == -1)
 		errExit("asprintf");
-	if (mkdir(odiff, S_IRWXU | S_IRWXG | S_IRWXO))
+	if (mkdir(odiff, 0755))
 		errExit("mkdir");
 	if (chown(odiff, 0, 0) < 0)
 		errExit("chown");
-	if (chmod(odiff, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
+	if (chmod(odiff, 0755) < 0)
 		errExit("chmod");
-
+	
 	char *owork;
-	if(asprintf(&owork, "%s/owork", MNT_DIR) == -1)
+	if(asprintf(&owork, "%s/owork", basedir) == -1)
 		errExit("asprintf");
-	if (mkdir(owork, S_IRWXU | S_IRWXG | S_IRWXO))
+	if (mkdir(owork, 0755))
 		errExit("mkdir");
 	if (chown(owork, 0, 0) < 0)
 		errExit("chown");
-	if (chmod(owork, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
+	if (chmod(owork, 0755) < 0)
 		errExit("chmod");
-
+	
 	// mount overlayfs
 	if (arg_debug)
 		printf("Mounting OverlayFS\n");
 	char *option;
 	if (oldkernel) { // old Ubuntu/OpenSUSE kernels
+		if (arg_overlay_keep) {
+			fprintf(stderr, "Error: option --overlay= not available for kernels older than 3.18\n");
+			exit(1);
+		}
 		if (asprintf(&option, "lowerdir=/,upperdir=%s", odiff) == -1)
 			errExit("asprintf");
 		if (mount("overlayfs", oroot, "overlayfs", MS_MGC_VAL, option) < 0)
@@ -666,10 +794,12 @@ void fs_overlayfs(void) {
 	else { // kernel 3.18 or newer
 		if (asprintf(&option, "lowerdir=/,upperdir=%s,workdir=%s", odiff, owork) == -1)
 			errExit("asprintf");
+//printf("option #%s#\n", option);			
 		if (mount("overlay", oroot, "overlay", MS_MGC_VAL, option) < 0)
 			errExit("mounting overlayfs");
 	}
-
+	printf("OverlayFS configured in %s directory\n", basedir);
+	
 	// mount-bind dev directory
 	if (arg_debug)
 		printf("Mounting /dev\n");
@@ -692,9 +822,10 @@ void fs_overlayfs(void) {
 	fs_var_cache();
 	fs_var_utmp();
 
-	// only in user mode
-	if (getuid())
-		sanitize_home();
+	// don't leak user information
+	restrict_users();
+
+	disable_firejail_config();
 
 	// cleanup and exit
 	free(option);
@@ -782,7 +913,7 @@ void fs_chroot(const char *rootdir) {
 	if (asprintf(&rundir, "%s/run", rootdir) == -1)
 		errExit("asprintf");
 	if (!is_dir(rundir)) {
-		int rv = mkdir(rundir, S_IRWXU | S_IRWXG | S_IRWXO);
+		int rv = mkdir(rundir, 0755);
 		(void) rv;
 		rv = chown(rundir, 0, 0);
 		(void) rv;
@@ -796,6 +927,10 @@ void fs_chroot(const char *rootdir) {
 		errExit("asprintf");
 	if (arg_debug)
 		printf("Updating /etc/resolv.conf in %s\n", fname);
+	if (is_link(fname)) {
+		fprintf(stderr, "Error: invalid %s file\n", fname);
+		exit(1);
+	}
 	if (copy_file("/etc/resolv.conf", fname) == -1)
 		fprintf(stderr, "Warning: /etc/resolv.conf not initialized\n");
 		
@@ -804,6 +939,8 @@ void fs_chroot(const char *rootdir) {
 		printf("Chrooting into %s\n", rootdir);
 	if (chroot(rootdir) < 0)
 		errExit("chroot");
+	// mount a new tmpfs in /run/firejail/mnt - the old one was lost in chroot
+	fs_build_remount_mnt_dir();
 		
 	// update /var directory in order to support multiple sandboxes running on the same root directory
 	if (!arg_private_dev)
@@ -815,11 +952,18 @@ void fs_chroot(const char *rootdir) {
 	fs_var_cache();
 	fs_var_utmp();
 
-	// only in user mode
-	if (getuid())
-		sanitize_home();
-	
+	// don't leak user information
+	restrict_users();
+
+	disable_firejail_config();
 }
 #endif
 
+void fs_private_tmp(void) {
+	// mount tmpfs on top of /run/firejail/mnt
+	if (arg_debug)
+		printf("Mounting tmpfs on /tmp directory\n");
+	if (mount("tmpfs", "/tmp", "tmpfs", MS_NOSUID | MS_STRICTATIME | MS_REC,  "mode=1777,gid=0") < 0)
+		errExit("mounting /tmp/firejail/mnt");
+}
 
