@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014, 2015 netblue30 (netblue30@yahoo.com)
+ * Copyright (C) 2014-2016 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -18,102 +18,9 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
 
-/* default seccomp filter
-	// seccomp
-	struct sock_filter filter[] = {
-		VALIDATE_ARCHITECTURE,
-		EXAMINE_SYSCALL,
-		BLACKLIST(SYS_mount),  // mount/unmount filesystems
-		BLACKLIST(SYS_umount2),
-		BLACKLIST(SYS_ptrace), // trace processes
-		BLACKLIST(SYS_kexec_load), // loading a different kernel
-		BLACKLIST(SYS_open_by_handle_at), // open by handle
-		BLACKLIST(SYS_init_module), // kernel module handling
-#ifdef SYS_finit_module // introduced in 2013
-		BLACKLIST(SYS_finit_module),
-#endif
-		BLACKLIST(SYS_delete_module),
-		BLACKLIST(SYS_iopl), // io permisions
-#ifdef SYS_ioperm
-		BLACKLIST(SYS_ioperm),
-#endif
-SYS_iopl
-		BLACKLIST(SYS_iopl), // io permisions
-#endif
-#ifdef SYS_ni_syscall), // new io permisions call on arm devices
-		BLACKLIST(SYS_ni_syscall),
-#endif
-		BLACKLIST(SYS_swapon), // swap on/off
-		BLACKLIST(SYS_swapoff),
-		BLACKLIST(SYS_syslog), // kernel printk control
-		RETURN_ALLOW
-	};
-*/
 #ifdef HAVE_SECCOMP
 #include "firejail.h"
-#include <errno.h>
-#include <linux/filter.h>
-#include <sys/syscall.h>
-#include <linux/capability.h>
-#include <linux/audit.h>
-#include <sys/stat.h>
-#include <fcntl.h>
-
-#include <sys/prctl.h>
-#ifndef PR_SET_NO_NEW_PRIVS
-# define PR_SET_NO_NEW_PRIVS 38
-#endif
-
-#if HAVE_SECCOMP_H
-#include <linux/seccomp.h>
-#else
-#define SECCOMP_MODE_FILTER	2
-#define SECCOMP_RET_KILL	0x00000000U
-#define SECCOMP_RET_TRAP	0x00030000U
-#define SECCOMP_RET_ALLOW	0x7fff0000U
-#define SECCOMP_RET_ERRNO	0x00050000U
-#define SECCOMP_RET_DATA        0x0000ffffU
-struct seccomp_data {
-    int nr;
-    __u32 arch;
-    __u64 instruction_pointer;
-    __u64 args[6];
-};
-#endif
-
-#if defined(__i386__)
-# define ARCH_NR	AUDIT_ARCH_I386
-#elif defined(__x86_64__)
-# define ARCH_NR	AUDIT_ARCH_X86_64
-#elif defined(__arm__)
-# define ARCH_NR	AUDIT_ARCH_ARM
-#else
-# warning "Platform does not support seccomp filter yet"
-# define ARCH_NR	0
-#endif
-
-
-#define VALIDATE_ARCHITECTURE \
-     BPF_STMT(BPF_LD+BPF_W+BPF_ABS, (offsetof(struct seccomp_data, arch))), \
-     BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ARCH_NR, 1, 0), \
-     BPF_STMT(BPF_RET+BPF_K, SECCOMP_RET_KILL)
-
-#define EXAMINE_SYSCALL BPF_STMT(BPF_LD+BPF_W+BPF_ABS,	\
-		 (offsetof(struct seccomp_data, nr)))
-
-#define BLACKLIST(syscall_nr)	\
-	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, syscall_nr, 0, 1),	\
-	BPF_STMT(BPF_RET+BPF_K, SECCOMP_RET_KILL)
-
-#define WHITELIST(syscall_nr) \
-	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, syscall_nr, 0, 1), \
-	BPF_STMT(BPF_RET+BPF_K, SECCOMP_RET_ALLOW)
-
-#define RETURN_ALLOW \
-	BPF_STMT(BPF_RET+BPF_K, SECCOMP_RET_ALLOW)
-
-#define KILL_PROCESS \
-	BPF_STMT(BPF_RET+BPF_K, SECCOMP_RET_KILL)
+#include "seccomp.h"
 
 #define SECSIZE 128 // initial filter size
 static struct sock_filter *sfilter = NULL;
@@ -157,6 +64,11 @@ void filter_debug(void) {
 			printf("  BLACKLIST %d %s\n", *nr, syscall_find_nr(*nr));
 			i += 2;
 		}
+		else if (*ptr	== 0x15 && *(ptr +14) == 0x5 && *(ptr + 15) == 0) {
+			int err = *(ptr + 13) << 8 | *(ptr + 12);
+			printf("  ERRNO %d %s %d %s\n", *nr, syscall_find_nr(*nr), err, errno_find_nr(err));
+			i += 2;
+		}
 		else if (*ptr == 0x06 && *(ptr +6) == 0 && *(ptr + 7) == 0 ) {
 			printf("  KILL_PROCESS\n");
 			i++;
@@ -179,8 +91,8 @@ static void filter_init(void) {
 		return;
 	}
 
-	if (arg_debug)
-		printf("Initialize seccomp filter\n");	
+//	if (arg_debug)
+//		printf("Initialize seccomp filter\n");	
 	// allocate a filter of SECSIZE
 	sfilter = malloc(sizeof(struct sock_filter) * SECSIZE);
 	if (!sfilter)
@@ -216,12 +128,13 @@ static void filter_realloc(void) {
 	sfilter_alloc_size += SECSIZE;
 }
 
-static void filter_add_whitelist(int syscall) {
+static void filter_add_whitelist(int syscall, int arg) {
+	(void) arg;
 	assert(sfilter);
 	assert(sfilter_alloc_size);
 	assert(sfilter_index);
-	if (arg_debug)
-		printf("Whitelisting syscall %d %s\n", syscall, syscall_find_nr(syscall));
+//	if (arg_debug)
+//		printf("Whitelisting syscall %d %s\n", syscall, syscall_find_nr(syscall));
 	
 	if ((sfilter_index + 2) > sfilter_alloc_size)
 		filter_realloc();
@@ -242,12 +155,13 @@ static void filter_add_whitelist(int syscall) {
 	sfilter_index += sizeof(filter) / sizeof(struct sock_filter);	
 }
 
-static void filter_add_blacklist(int syscall) {
+static void filter_add_blacklist(int syscall, int arg) {
+	(void) arg;
 	assert(sfilter);
 	assert(sfilter_alloc_size);
 	assert(sfilter_index);
-	if (arg_debug)
-		printf("Blacklisting syscall %d %s\n", syscall, syscall_find_nr(syscall));
+//	if (arg_debug)
+//		printf("Blacklisting syscall %d %s\n", syscall, syscall_find_nr(syscall));
 	
 	if ((sfilter_index + 2) > sfilter_alloc_size)
 		filter_realloc();
@@ -268,12 +182,38 @@ static void filter_add_blacklist(int syscall) {
 	sfilter_index += sizeof(filter) / sizeof(struct sock_filter);	
 }
 
+static void filter_add_errno(int syscall, int arg) {
+	assert(sfilter);
+	assert(sfilter_alloc_size);
+	assert(sfilter_index);
+//	if (arg_debug)
+//		printf("Errno syscall %d %d %s\n", syscall, arg, syscall_find_nr(syscall));
+
+	if ((sfilter_index + 2) > sfilter_alloc_size)
+		filter_realloc();
+
+	struct sock_filter filter[] = {
+		BLACKLIST_ERRNO(syscall, arg)
+	};
+#if 0
+{
+	int i;
+	unsigned char *ptr = (unsigned char *) &filter[0];
+	for (i = 0; i < sizeof(filter); i++, ptr++)
+		printf("%x, ", (*ptr) & 0xff);
+	printf("\n");
+}
+#endif
+	memcpy(&sfilter[sfilter_index], filter, sizeof(filter));
+	sfilter_index += sizeof(filter) / sizeof(struct sock_filter);
+}
+
 static void filter_end_blacklist(void) {
 	assert(sfilter);
 	assert(sfilter_alloc_size);
 	assert(sfilter_index);
-	if (arg_debug)
-		printf("Ending syscall filter\n");
+//	if (arg_debug)
+//		printf("Ending syscall filter\n");
 
 	if ((sfilter_index + 2) > sfilter_alloc_size)
 		filter_realloc();
@@ -326,39 +266,27 @@ static void write_seccomp_file(void) {
 	fs_build_mnt_dir();
 	assert(sfilter);
 
-	char *fname;
-	if (asprintf(&fname, "%s/seccomp", MNT_DIR) == -1)
-		errExit("asprintf");
-	int fd = open(fname, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR);
+	int fd = open(RUN_SECCOMP_CFG, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR);
 	if (fd == -1)
 		errExit("open");
 
 	if (arg_debug)
-		printf("Save seccomp filter, size %lu bytes\n", sfilter_index * sizeof(struct sock_filter));
+		printf("Save seccomp filter, size %u bytes\n", (unsigned) (sfilter_index * sizeof(struct sock_filter)));
 	errno = 0;
 	ssize_t sz = write(fd, sfilter, sfilter_index * sizeof(struct sock_filter));
-	if (sz != (sfilter_index * sizeof(struct sock_filter))) {
+	if (sz != (ssize_t)(sfilter_index * sizeof(struct sock_filter))) {
 		fprintf(stderr, "Error: cannot save seccomp filter\n");
 		exit(1);
 	}
 	close(fd);
-	if (chown(fname, 0, 0) < 0)
+	if (chown(RUN_SECCOMP_CFG, 0, 0) < 0)
 		errExit("chown");
-	free(fname);
 }
 
 // read seccomp filter from /tmp/firejail/mnt/seccomp
-static void read_seccomp_file(char *file_name) {
+static void read_seccomp_file(const char *fname) {
 	assert(sfilter == NULL && sfilter_index == 0);
 
-	char *fname;
-	if (file_name)
-		fname = file_name;
-	else {
-		if (asprintf(&fname, "%s/seccomp", MNT_DIR) == -1)
-			errExit("asprintf");
-	}
-		
 	// check file
 	struct stat s;
 	if (stat(fname, &s) == -1) {
@@ -388,111 +316,264 @@ static void read_seccomp_file(char *file_name) {
 	sfilter_index = sz / sizeof(struct sock_filter);
 
 	if (arg_debug)
-		printf("Read seccomp filter, size %lu bytes\n", sfilter_index * sizeof(struct sock_filter));
+		printf("Read seccomp filter, size %u bytes\n", (unsigned) (sfilter_index * sizeof(struct sock_filter)));
 
 	close(fd);
-	free(fname);
 	
 	if (arg_debug)
 		filter_debug();
 }
 
+// i386t filter installed on amd64 architectures
+void seccomp_filter_32(void) {
+	// hardcoded syscall values
+	struct sock_filter filter[] = {
+		VALIDATE_ARCHITECTURE_32,
+		EXAMINE_SYSCALL,
+		BLACKLIST(21), // mount
+		BLACKLIST(52), // umount2
+		BLACKLIST(26), // ptrace
+		BLACKLIST(283), // kexec_load
+		BLACKLIST(342), // open_by_handle_at
+		BLACKLIST(128), // init_module
+		BLACKLIST(350), // finit_module
+		BLACKLIST(129), // delete_module
+		BLACKLIST(110), // iopl
+		BLACKLIST(101), // ioperm
+		BLACKLIST(87), // swapon
+		BLACKLIST(115), // swapoff
+		BLACKLIST(103), // syslog
+		BLACKLIST(347), // process_vm_readv
+		BLACKLIST(348), // process_vm_writev
+		BLACKLIST(135), // sysfs
+		BLACKLIST(149), // _sysctl
+		BLACKLIST(124), // adjtimex
+		BLACKLIST(343), // clock_adjtime
+		BLACKLIST(253), // lookup_dcookie
+		BLACKLIST(336), // perf_event_open
+		BLACKLIST(338), // fanotify_init
+		BLACKLIST(349), // kcmp
+		BLACKLIST(286), // add_key
+		BLACKLIST(287), // request_key
+		BLACKLIST(288), // keyctl
+		BLACKLIST(86), // uselib
+		BLACKLIST(51), // acct
+		BLACKLIST(123), // modify_ldt
+		BLACKLIST(217), // pivot_root
+		BLACKLIST(245), // io_setup
+		BLACKLIST(246), // io_destroy
+		BLACKLIST(247), // io_getevents
+		BLACKLIST(248), // io_submit
+		BLACKLIST(249), // io_cancel
+		BLACKLIST(257), // remap_file_pages
+		BLACKLIST(274), // mbind
+		BLACKLIST(275), // get_mempolicy
+		BLACKLIST(276), // set_mempolicy
+		BLACKLIST(294), // migrate_pages
+		BLACKLIST(317), // move_pages
+		BLACKLIST(316), // vmsplice
+		BLACKLIST(61),  // chroot
+		RETURN_ALLOW
+	};
+
+	struct sock_fprog prog = {
+		.len = (unsigned short)(sizeof(filter) / sizeof(filter[0])),
+		.filter = filter,
+	};
+
+	if (prctl(PR_SET_SECCOMP, SECCOMP_MODE_FILTER, &prog) || prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0)) {
+		;
+	}
+	else if (arg_debug) {
+		printf("Dual i386/amd64 seccomp filter configured\n");
+	}
+}
 
 // drop filter for seccomp option
-int seccomp_filter_drop(void) {
+int seccomp_filter_drop(int enforce_seccomp) {
 	filter_init();
 	
 	// default seccomp
-	if (arg_seccomp_list_drop == NULL) {
+	if (cfg.seccomp_list_drop == NULL) {
+#if defined(__x86_64__)
+		seccomp_filter_32();
+#endif
+
 #ifdef SYS_mount		
-		filter_add_blacklist(SYS_mount);
+		filter_add_blacklist(SYS_mount, 0);
 #endif
 #ifdef SYS_umount2		
-		filter_add_blacklist(SYS_umount2);
+		filter_add_blacklist(SYS_umount2, 0);
 #endif
 #ifdef SYS_ptrace 		
-		filter_add_blacklist(SYS_ptrace);
+		filter_add_blacklist(SYS_ptrace, 0);
 #endif
 #ifdef SYS_kexec_load		
-		filter_add_blacklist(SYS_kexec_load);
+		filter_add_blacklist(SYS_kexec_load, 0);
+#endif
+#ifdef SYS_kexec_file_load		
+		filter_add_blacklist(SYS_kexec_file_load, 0);
 #endif
 #ifdef SYS_open_by_handle_at		
-		filter_add_blacklist(SYS_open_by_handle_at);
+		filter_add_blacklist(SYS_open_by_handle_at, 0);
 #endif
 #ifdef SYS_init_module		
-		filter_add_blacklist(SYS_init_module);
+		filter_add_blacklist(SYS_init_module, 0);
 #endif
 #ifdef SYS_finit_module // introduced in 2013
-		filter_add_blacklist(SYS_finit_module);
+		filter_add_blacklist(SYS_finit_module, 0);
 #endif
 #ifdef SYS_delete_module		
-		filter_add_blacklist(SYS_delete_module);
+		filter_add_blacklist(SYS_delete_module, 0);
 #endif
 #ifdef SYS_iopl		
-		filter_add_blacklist(SYS_iopl);
+		filter_add_blacklist(SYS_iopl, 0);
 #endif
 #ifdef 	SYS_ioperm	
-		filter_add_blacklist(SYS_ioperm);
+		filter_add_blacklist(SYS_ioperm, 0);
 #endif
-#ifdef SYS_ni_syscall // new io permisions call on arm devices
-		filter_add_blacklist(SYS_ni_syscall);
+#ifdef SYS_ni_syscall // new io permissions call on arm devices
+		filter_add_blacklist(SYS_ni_syscall, 0);
 #endif
 #ifdef SYS_swapon		
-		filter_add_blacklist(SYS_swapon);
+		filter_add_blacklist(SYS_swapon, 0);
 #endif
 #ifdef SYS_swapoff		
-		filter_add_blacklist(SYS_swapoff);
+		filter_add_blacklist(SYS_swapoff, 0);
 #endif
 #ifdef SYS_syslog		
-		filter_add_blacklist(SYS_syslog);
+		filter_add_blacklist(SYS_syslog, 0);
 #endif
 #ifdef SYS_process_vm_readv		
-		filter_add_blacklist(SYS_process_vm_readv);
+		filter_add_blacklist(SYS_process_vm_readv, 0);
 #endif
 #ifdef SYS_process_vm_writev		
-		filter_add_blacklist(SYS_process_vm_writev);
-#endif
-#ifdef SYS_mknod		
-		filter_add_blacklist(SYS_mknod);
+		filter_add_blacklist(SYS_process_vm_writev, 0);
 #endif
+
+// mknod removed in 0.9.29 - it brakes Zotero extension
+//#ifdef SYS_mknod		
+//		filter_add_blacklist(SYS_mknod, 0);
+//#endif
 		
 		// new syscalls in 0.9,23		
 #ifdef SYS_sysfs		
-		filter_add_blacklist(SYS_sysfs);
+		filter_add_blacklist(SYS_sysfs, 0);
 #endif
 #ifdef SYS__sysctl	
-		filter_add_blacklist(SYS__sysctl);
+		filter_add_blacklist(SYS__sysctl, 0);
 #endif
 #ifdef SYS_adjtimex		
-		filter_add_blacklist(SYS_adjtimex);
+		filter_add_blacklist(SYS_adjtimex, 0);
 #endif
 #ifdef 	SYS_clock_adjtime	
-		filter_add_blacklist(SYS_clock_adjtime);
+		filter_add_blacklist(SYS_clock_adjtime, 0);
 #endif
 #ifdef SYS_lookup_dcookie		
-		filter_add_blacklist(SYS_lookup_dcookie);
+		filter_add_blacklist(SYS_lookup_dcookie, 0);
 #endif
 #ifdef 	SYS_perf_event_open	
-		filter_add_blacklist(SYS_perf_event_open);
+		filter_add_blacklist(SYS_perf_event_open, 0);
 #endif
 #ifdef	SYS_fanotify_init 	
-		filter_add_blacklist(SYS_fanotify_init);
+		filter_add_blacklist(SYS_fanotify_init, 0);
 #endif
 #ifdef SYS_kcmp
-		filter_add_blacklist(SYS_kcmp);
+		filter_add_blacklist(SYS_kcmp, 0);
 #endif
+
+// 0.9.32
+#ifdef SYS_add_key
+		filter_add_blacklist(SYS_add_key, 0);
+#endif
+#ifdef SYS_request_key
+		filter_add_blacklist(SYS_request_key, 0);
+#endif
+#ifdef SYS_keyctl
+		filter_add_blacklist(SYS_keyctl, 0);
+#endif
+#ifdef SYS_uselib
+		filter_add_blacklist(SYS_uselib, 0);
+#endif
+#ifdef SYS_acct
+		filter_add_blacklist(SYS_acct, 0);
+#endif
+#ifdef SYS_modify_ldt
+		filter_add_blacklist(SYS_modify_ldt, 0);
+#endif
+	//#ifdef SYS_unshare
+	//		filter_add_blacklist(SYS_unshare, 0);
+	//#endif
+#ifdef SYS_pivot_root
+		filter_add_blacklist(SYS_pivot_root, 0);
+#endif
+	//#ifdef SYS_quotactl
+	//		filter_add_blacklist(SYS_quotactl, 0);
+	//#endif
+#ifdef SYS_io_setup
+		filter_add_blacklist(SYS_io_setup, 0);
+#endif
+#ifdef SYS_io_destroy
+		filter_add_blacklist(SYS_io_destroy, 0);
+#endif
+#ifdef SYS_io_getevents
+		filter_add_blacklist(SYS_io_getevents, 0);
+#endif
+#ifdef SYS_io_submit
+		filter_add_blacklist(SYS_io_submit, 0);
+#endif
+#ifdef SYS_io_cancel
+		filter_add_blacklist(SYS_io_cancel, 0);
+#endif
+#ifdef SYS_remap_file_pages
+		filter_add_blacklist(SYS_remap_file_pages, 0);
+#endif
+#ifdef SYS_mbind
+		filter_add_blacklist(SYS_mbind, 0);
+#endif
+#ifdef SYS_get_mempolicy
+		filter_add_blacklist(SYS_get_mempolicy, 0);
+#endif
+#ifdef SYS_set_mempolicy
+		filter_add_blacklist(SYS_set_mempolicy, 0);
+#endif
+#ifdef SYS_migrate_pages
+		filter_add_blacklist(SYS_migrate_pages, 0);
+#endif
+#ifdef SYS_move_pages
+		filter_add_blacklist(SYS_move_pages, 0);
+#endif
+#ifdef SYS_vmsplice
+		filter_add_blacklist(SYS_vmsplice, 0);
+#endif
+#ifdef SYS_chroot
+		filter_add_blacklist(SYS_chroot, 0);
+#endif
+	//#ifdef SYS_set_robust_list
+	//		filter_add_blacklist(SYS_set_robust_list, 0);
+	//#endif
+	//#ifdef SYS_get_robust_list
+	//		filter_add_blacklist(SYS_get_robust_list, 0);
+	//#endif
+
+ // CHECK_SECCOMP(seccomp_rule_add(ctx, SCMP_ACT_ERRNO(EPERM), SCMP_SYS(clone), 1,
+ //     SCMP_A0(SCMP_CMP_MASKED_EQ, CLONE_NEWUSER, CLONE_NEWUSER)));
+
+// 32bit
+//		filter_add_blacklist(SYS_personality, 0); // test wine
+//		filter_add_blacklist(SYS_set_thread_area, 0); // test wine
 	}
 
 	// default seccomp filter with additional drop list
-	if (arg_seccomp_list && arg_seccomp_list_drop == NULL) {
-		if (syscall_check_list(arg_seccomp_list, filter_add_blacklist)) {
+	if (cfg.seccomp_list && cfg.seccomp_list_drop == NULL) {
+		if (syscall_check_list(cfg.seccomp_list, filter_add_blacklist, 0)) {
 			fprintf(stderr, "Error: cannot load seccomp filter\n");
 			exit(1);
 		}
 	}
 	// drop list
-	else if (arg_seccomp_list == NULL && arg_seccomp_list_drop) {
-		if (syscall_check_list(arg_seccomp_list_drop, filter_add_blacklist)) {
+	else if (cfg.seccomp_list == NULL && cfg.seccomp_list_drop) {
+		if (syscall_check_list(cfg.seccomp_list_drop, filter_add_blacklist, 0)) {
 			fprintf(stderr, "Error: cannot load seccomp filter\n");
 			exit(1);
 		}
@@ -514,12 +595,15 @@ int seccomp_filter_drop(void) {
 	};
 
 	if (prctl(PR_SET_SECCOMP, SECCOMP_MODE_FILTER, &prog) || prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0)) {
-		fprintf(stderr, "Warning: seccomp disabled, it requires a Linux kernel version 3.5 or newer.\n");
+		if (enforce_seccomp) {
+			fprintf(stderr, "Error: a seccomp-enabled Linux kernel is required, exiting...\n");
+			exit(1);
+		}
+		else
+			fprintf(stderr, "Warning: seccomp disabled, it requires a Linux kernel version 3.5 or newer.\n");
+
 		return 1;
 	}
-	else if (arg_debug) {
-		printf("seccomp enabled\n");
-	}
 	
 	return 0;
 }
@@ -529,14 +613,14 @@ int seccomp_filter_keep(void) {
 	filter_init();
 
 	// these 4 syscalls are used by firejail after the seccomp filter is initialized
-	filter_add_whitelist(SYS_setuid);
-	filter_add_whitelist(SYS_setgid);
-	filter_add_whitelist(SYS_setgroups);
-	filter_add_whitelist(SYS_dup);
+	filter_add_whitelist(SYS_setuid, 0);
+	filter_add_whitelist(SYS_setgid, 0);
+	filter_add_whitelist(SYS_setgroups, 0);
+	filter_add_whitelist(SYS_dup, 0);
 	
 	// apply keep list
-	if (arg_seccomp_list_keep) {
-		if (syscall_check_list(arg_seccomp_list_keep, filter_add_whitelist)) {
+	if (cfg.seccomp_list_keep) {
+		if (syscall_check_list(cfg.seccomp_list_keep, filter_add_whitelist, 0)) {
 			fprintf(stderr, "Error: cannot load seccomp filter\n");
 			exit(1);
 		}
@@ -567,11 +651,52 @@ int seccomp_filter_keep(void) {
 	return 0;
 }
 
+// errno filter for seccomp option
+int seccomp_filter_errno(void) {
+	int i;
+	int higest_errno = errno_highest_nr();
+	filter_init();
+
+	// apply errno list
+
+	for (i = 0; i < higest_errno; i++) {
+		if (cfg.seccomp_list_errno[i]) {
+			if (syscall_check_list(cfg.seccomp_list_errno[i], filter_add_errno, i)) {
+				fprintf(stderr, "Error: cannot load seccomp filter\n");
+				exit(1);
+			}
+		}
+	}
+
+	filter_end_blacklist();
+	if (arg_debug)
+		filter_debug();
+
+	// save seccomp filter in  /tmp/firejail/mnt/seccomp
+	// in order to use it in --join operations
+	write_seccomp_file();
+
+	struct sock_fprog prog = {
+		.len = sfilter_index,
+		.filter = sfilter,
+	};
+
+	if (prctl(PR_SET_SECCOMP, SECCOMP_MODE_FILTER, &prog) || prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0)) {
+		fprintf(stderr, "Warning: seccomp disabled, it requires a Linux kernel version 3.5 or newer.\n");
+		return 1;
+	}
+	else if (arg_debug) {
+		printf("seccomp enabled\n");
+	}
+
+	return 0;
+}
+
 
 
 void seccomp_set(void) {
 	// read seccomp filter from  /tmp/firejail/mnt/seccomp
-	read_seccomp_file(NULL);
+	read_seccomp_file(RUN_SECCOMP_CFG);
 	
 	// apply filter
 	struct sock_fprog prog = {
@@ -622,14 +747,9 @@ void seccomp_print_filter(pid_t pid) {
 	// check privileges for non-root users
 	uid_t uid = getuid();
 	if (uid != 0) {
-		struct stat s;
-		char *dir;
-		if (asprintf(&dir, "/proc/%u/ns", pid) == -1)
-			errExit("asprintf");
-		if (stat(dir, &s) < 0)
-			errExit("stat");
-		if (s.st_uid != uid) {
-			printf("Error: permission denied.\n");
+		uid_t sandbox_uid = pid_get_uid(pid);
+		if (uid != sandbox_uid) {
+			fprintf(stderr, "Error: permission denied.\n");
 			exit(1);
 		}
 	}
@@ -637,7 +757,7 @@ void seccomp_print_filter(pid_t pid) {
 
 	// find the seccomp filter
 	char *fname;
-	if (asprintf(&fname, "/proc/%d/root/tmp/firejail/mnt/seccomp", pid) == -1)
+	if (asprintf(&fname, "/proc/%d/root%s", pid, RUN_SECCOMP_CFG) == -1)
 		errExit("asprintf");
 
 	struct stat s;
@@ -650,6 +770,7 @@ void seccomp_print_filter(pid_t pid) {
 	read_seccomp_file(fname);
 	drop_privs(1);
 	filter_debug();
+	free(fname);
 
 	exit(0);
 }
